"""Top-level package for piconetcontrol."""

__author__ = """Matthias Gilles Zeller"""
__author_email__ = "matthias.gilles.zeller@gmail.com"
__version__ = "0.2.1"
__description__ = "Client-server package to remotely control a Raspberrypi Pi Pico W."
__url__ = "https://github.com/matthiaszeller/picoserver"
